###
# Copyright (c) 2016, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

###

import supybot.utils as utils
from supybot.commands import *
import supybot.plugins as plugins
import supybot.ircutils as ircutils
import supybot.callbacks as callbacks
try:
    from supybot.i18n import PluginInternationalization
    _ = PluginInternationalization('Lists')
except ImportError:
    # Placeholder that allows to run the plugin on a bot
    # without the i18n module
    _ = lambda x: x


class Lists(callbacks.Plugin):
    """Utilities for dealing with lists."""
    
    @wrap(['int', optional('int'), optional('int')])
    def range(self, irc, msg, args, arg1, arg2, step):
        """[start] end [step]

        Returns a range, from start (defaults to 0) to end, using a
        step (defaults to 1)."""
        if arg2 is None:
            (start, end) = (0, arg1)
        else:
            (start, end) = (arg1, arg2)
        if step is None:
            step = 1
        maximum = self.registryValue('range.maximum')
        minimum = self.registryValue('range.minimum')
        if end > maximum or start > maximum:
            irc.error(_('Range contains values larger than allowed by '
                'supybot.plugins.Lists.range.maximum'), Raise=True)
        if end < minimum or end < minimum:
            irc.error(_('Range contains values larger than allowed by '
                'supybot.plugins.Lists.range.minimum'), Raise=True)
        if step == 0:
            irc.error(_('Step cannot be zero.'), Raise=True)
        if (end - start)/step > self.registryValue('range.maximumLength'):
            irc.error(_('Range is longer than allowed by '
                'supybot.plugins.Lists.range.maximumLength'), Raise=True)
        irc.replies(list(map(str, range(start, end, step))))

    @wrap(['something', 'something'])


Class = Lists


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2002-2005, Jeremiah Fincher
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from supybot.test import *

import sys

import supybot.world as world

class StatusTestCase(PluginTestCase):
    plugins = ('Status',)
    def testNet(self):
        self.assertNotError('net')

    def testCpu(self):
        m = self.assertNotError('status cpu')
        self.failIf('kB kB' in m.args[1])
        self.failIf('None' in m.args[1], 'None in cpu output: %r.' % m)
        for s in ['linux', 'freebsd', 'openbsd', 'netbsd', 'darwin']:
            if sys.platform.startswith(s):
                self.failUnless('B' in m.args[1] or 'KB' in m.args[1] or
                                'MB' in m.args[1],
                                'No memory string on supported platform.')
        try:
            original = conf.supybot.plugins.Status.cpu.get('children')()
            conf.supybot.plugins.Status.cpu.get('children').setValue(False)
            self.assertNotRegexp('cpu', 'children')
        finally:
            conf.supybot.plugins.Status.cpu.get('children').setValue(original)
            

    def testUptime(self):
        self.assertNotError('uptime')

    def testCmd(self):
        self.assertNotError('cmd')

    def testCommands(self):
        self.assertNotError('commands')

    def testLogfilesize(self):
        self.feedMsg('list')
        self.feedMsg('list Status')
        self.assertNotError('upkeep')

    def testThreads(self):
        self.assertNotError('threads')

    def testProcesses(self):
        self.assertNotError('processes')

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:




###
# Copyright (c) 2005, Jeremiah Fincher
# Copyright (c) 2009-2010, James McCoy
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import supybot.conf as conf
import supybot.registry as registry
from supybot.i18n import PluginInternationalization, internationalizeDocstring
_ = PluginInternationalization('ShrinkUrl')

def configure(advanced):
    from supybot.questions import output, expect, anything, something, yn
    conf.registerPlugin('ShrinkUrl', True)
    if yn(_("""This plugin offers a snarfer that will go retrieve a shorter
             version of long URLs that are sent to the channel.  Would you
             like this snarfer to be enabled?"""), default=False):
        conf.supybot.plugins.ShrinkUrl.shrinkSnarfer.setValue(True)

class ShrinkService(registry.OnlySomeStrings):
    """Valid values include 'tiny', 'ur1', and 'x0'."""
    validStrings = ('tiny', 'ur1', 'x0')

class ShrinkCycle(registry.SpaceSeparatedListOfStrings):
    """Valid values include 'ln', 'tiny', 'ur1', and 'x0'."""
    Value = ShrinkService

    def __init__(self, *args, **kwargs):
        super(ShrinkCycle, self).__init__(*args, **kwargs)
        self.lastIndex = -1

    def setValue(self, v):
        super(self.__class__, self).setValue(v)
        self.lastIndex = -1

    def getService(self):
        L = self()
        if L:
            self.lastIndex = (self.lastIndex + 1) % len(L)
            return L[self.lastIndex]
        raise ValueError('No services have been configured for rotation.  ' \
                'See conf.supybot.plugins.ShrinkUrl.serviceRotation.')

ShrinkUrl = conf.registerPlugin('ShrinkUrl')
conf.registerChannelValue(ShrinkUrl, 'shrinkSnarfer',
    registry.Boolean(False, _("""Determines whether the
    shrink snarfer is enabled.  This snarfer will watch for URLs in the
    channel, and if they're sufficiently long (as determined by
    supybot.plugins.ShrinkUrl.minimumLength) it will post a
    smaller URL from the service as denoted in
    supybot.plugins.ShrinkUrl.default.""")))
conf.registerChannelValue(ShrinkUrl.shrinkSnarfer, 'showDomain',
    registry.Boolean(True, _("""Determines whether the snarfer will show the
    domain of the URL being snarfed along with the shrunken URL.""")))
conf.registerChannelValue(ShrinkUrl, 'minimumLength',
    registry.PositiveInteger(48, _("""The minimum length a URL must be before
    the bot will shrink it.""")))
conf.registerChannelValue(ShrinkUrl, 'nonSnarfingRegexp',
    registry.Regexp(None, _("""Determines what URLs are to be snarfed; URLs
    matching the regexp given will not be snarfed.  Give the empty string if
    you have no URLs that you'd like to exclude from being snarfed.""")))
conf.registerChannelValue(ShrinkUrl, 'outFilter',
    registry.Boolean(False, _("""Determines whether the bot will shrink the
    URLs of outgoing messages if those URLs are longer than
    supybot.plugins.ShrinkUrl.minimumLength.""")))
conf.registerChannelValue(ShrinkUrl, 'default',
    ShrinkService('x0', _("""Determines what website the bot will use when
    shrinking a URL.""")))
conf.registerGlobalValue(ShrinkUrl, 'bold',
    registry.Boolean(True, _("""Determines whether this plugin will bold
    certain portions of its replies.""")))
conf.registerChannelValue(ShrinkUrl, 'serviceRotation',
    ShrinkCycle([], _("""If set to a non-empty value, specifies the list of
    services to rotate through for the shrinkSnarfer and outFilter.""")))

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2003, Brett Kelly
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from supybot.test import *

class NoteTestCase(PluginTestCase):
    plugins = ('Note', 'Misc', 'User')
    config = {'supybot.reply.whenNotCommand': False}
    def setUp(self):
        PluginTestCase.setUp(self)
        # setup a user
        self.prefix = 'foo!bar@baz'
        self.assertNotError('register inkedmn bar')
        self.assertNotError('hostmask add inkedmn test2!bar@baz')

    def testSendnote(self):
        self.assertRegexp('note send inkedmn test', '#1')
        # have to getMsg(' ') after each Note.send to absorb supybot's
        # automatic "You have an unread note" message
        _ = self.getMsg(' ')
        self.assertError('note send alsdkjfasldk foo')
        self.assertNotError('note send inkedmn test2')
        _ = self.getMsg(' ')
        # verify that sending a note to a user via their nick instead of their
        # ircdb user name works
        self.prefix = 'test2!bar@baz'
        self.assertNotError('note send test2 foo')
        _ = self.getMsg(' ')

    def testNote(self):
        self.assertNotError('note send inkedmn test')
        _ = self.getMsg(' ')
        self.assertRegexp('note 1', 'test')
        self.assertError('note blah')

    def testList(self):
        self.assertResponse('note list', 'You have no unread notes.')
        self.assertNotError('note send inkedmn testing')
        _ = self.getMsg(' ')
        self.assertNotError('note send inkedmn 1,2,3')
        _ = self.getMsg(' ')
        self.assertRegexp('note list --sent', r'#2.*#1')
        self.assertRegexp('note list --sent --to inkedmn', r'#2.*#1')
        self.assertRegexp('note list', r'#1.*#2')
        self.assertRegexp('note 1', 'testing')
        self.assertRegexp('note list --old', '#1 from inkedmn')
        self.assertRegexp('note list --old --from inkedmn','#1 from inkedmn')

    def testSearch(self):
        self.assertNotError('note send inkedmn testing')
        _ = self.getMsg(' ')
        self.assertNotError('note send inkedmn 1,2,3')
        _ = self.getMsg(' ')
        self.assertRegexp('note search test', r'#1')
        self.assertRegexp('note search --regexp m/1,2/', r'#2')
        self.assertRegexp('note search --sent test', r'#1')

    def testNext(self):
        self.assertNotError('note send inkedmn testing')
        _ = self.getMsg(' ')
        self.assertRegexp('note next', 'testing')


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



# -*- coding: utf8 -*-
###
# Copyright (c) 2002-2005, Jeremiah Fincher
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from supybot.test import *

import supybot.conf as conf
import supybot.utils as utils
import supybot.ircmsgs as ircmsgs
import supybot.utils.minisix as minisix
import supybot.callbacks as callbacks

tokenize = callbacks.tokenize


class TokenizerTestCase(SupyTestCase):
    def testEmpty(self):
        self.assertEqual(tokenize(''), [])

    def testNullCharacter(self):
        self.assertEqual(tokenize(utils.str.dqrepr('\0')), ['\0'])

    def testSingleDQInDQString(self):
        self.assertEqual(tokenize('"\\""'), ['"'])

    def testDQsWithBackslash(self):
        self.assertEqual(tokenize('"\\\\"'), ["\\"])

    def testDoubleQuotes(self):
        self.assertEqual(tokenize('"\\"foo\\""'), ['"foo"'])

    def testSingleWord(self):
        self.assertEqual(tokenize('foo'), ['foo'])

    def testMultipleSimpleWords(self):
        words = 'one two three four five six seven eight'.split()
        for i in range(len(words)):
            self.assertEqual(tokenize(' '.join(words[:i])), words[:i])

    def testSingleQuotesNotQuotes(self):
        self.assertEqual(tokenize("it's"), ["it's"])

    def testQuotedWords(self):
        self.assertEqual(tokenize('"foo bar"'), ['foo bar'])
        self.assertEqual(tokenize('""'), [''])
        self.assertEqual(tokenize('foo "" bar'), ['foo', '', 'bar'])
        self.assertEqual(tokenize('foo "bar baz" quux'),
                         ['foo', 'bar baz', 'quux'])

    _testUnicode = """
def testUnicode(self):
    self.assertEqual(tokenize(u'好'), [u'好'])
    self.assertEqual(tokenize(u'"好"'), [u'好'])"""
    if minisix.PY3:
        _testUnicode = _testUnicode.replace("u'", "'")
    exec(_testUnicode)

    def testNesting(self):
        self.assertEqual(tokenize('[]'), [[]])
        self.assertEqual(tokenize('[foo]'), [['foo']])
        self.assertEqual(tokenize('[ foo ]'), [['foo']])
        self.assertEqual(tokenize('foo [bar]'), ['foo', ['bar']])
        self.assertEqual(tokenize('foo bar [baz quux]'),
                         ['foo', 'bar', ['baz', 'quux']])
        try:
            orig = conf.supybot.commands.nested()
            conf.supybot.commands.nested.setValue(False)
            self.assertEqual(tokenize('[]'), ['[]'])
            self.assertEqual(tokenize('[foo]'), ['[foo]'])
            self.assertEqual(tokenize('foo [bar]'), ['foo', '[bar]'])
            self.assertEqual(tokenize('foo bar [baz quux]'),
                             ['foo', 'bar', '[baz', 'quux]'])
        finally:
            conf.supybot.commands.nested.setValue(orig)

    def testError(self):
        self.assertRaises(SyntaxError, tokenize, '[foo') #]
        self.assertRaises(SyntaxError, tokenize, '"foo') #"

    def testPipe(self):
        try:
            conf.supybot.commands.nested.pipeSyntax.setValue(True)
            self.assertRaises(SyntaxError, tokenize, '| foo')
            self.assertRaises(SyntaxError, tokenize, 'foo ||bar')
            self.assertRaises(SyntaxError, tokenize, 'bar |')
            self.assertEqual(tokenize('foo|bar'), ['bar', ['foo']])
            self.assertEqual(tokenize('foo | bar'), ['bar', ['foo']])
            self.assertEqual(tokenize('foo | bar | baz'),
                             ['baz', ['bar',['foo']]])
            self.assertEqual(tokenize('foo bar | baz'),
                             ['baz', ['foo', 'bar']])
            self.assertEqual(tokenize('foo | bar baz'),
                             ['bar', 'baz', ['foo']])
            self.assertEqual(tokenize('foo bar | baz quux'),
                             ['baz', 'quux', ['foo', 'bar']])
        finally:
            conf.supybot.commands.nested.pipeSyntax.setValue(False)
            self.assertEqual(tokenize('foo|bar'), ['foo|bar'])
            self.assertEqual(tokenize('foo | bar'), ['foo', '|', 'bar'])
            self.assertEqual(tokenize('foo | bar | baz'),
                             ['foo', '|', 'bar', '|', 'baz'])
            self.assertEqual(tokenize('foo bar | baz'),
                             ['foo', 'bar', '|', 'baz'])

    def testQuoteConfiguration(self):
        f = callbacks.tokenize
        self.assertEqual(f('[foo]'), [['foo']])
        self.assertEqual(f('"[foo]"'), ['[foo]'])
        try:
            original = conf.supybot.commands.quotes()
            conf.supybot.commands.quotes.setValue('`')
            self.assertEqual(f('[foo]'), [['foo']])
            self.assertEqual(f('`[foo]`'), ['[foo]'])
            conf.supybot.commands.quotes.setValue('\'')
            self.assertEqual(f('[foo]'), [['foo']])
            self.assertEqual(f('\'[foo]\''), ['[foo]'])
            conf.supybot.commands.quotes.setValue('`\'')
            self.assertEqual(f('[foo]'), [['foo']])
            self.assertEqual(f('`[foo]`'), ['[foo]'])
            self.assertEqual(f('[foo]'), [['foo']])
            self.assertEqual(f('\'[foo]\''), ['[foo]'])
        finally:
            conf.supybot.commands.quotes.setValue(original)

    def testBold(self):
        s = '\x02foo\x02'
        self.assertEqual(tokenize(s), [s])
        s = s[:-1] + '\x0f'
        self.assertEqual(tokenize(s), [s])

    def testColor(self):
        s = '\x032,3foo\x03'
        self.assertEqual(tokenize(s), [s])
        s = s[:-1] + '\x0f'
        self.assertEqual(tokenize(s), [s])


class FunctionsTestCase(SupyTestCase):
    def testCanonicalName(self):
        self.assertEqual('foo', callbacks.canonicalName('foo'))
        self.assertEqual('foobar', callbacks.canonicalName('foo-bar'))
        self.assertEqual('foobar', callbacks.canonicalName('foo_bar'))
        self.assertEqual('foobar', callbacks.canonicalName('FOO-bar'))
        self.assertEqual('foobar', callbacks.canonicalName('FOOBAR'))
        self.assertEqual('foobar', callbacks.canonicalName('foo___bar'))
        self.assertEqual('foobar', callbacks.canonicalName('_f_o_o-b_a_r'))
        # The following seems to be a hack for the Karma plugin; I'm not
        # entirely sure that it's completely necessary anymore.
        self.assertEqual('foobar--', callbacks.canonicalName('foobar--'))

    def testAddressed(self):
        oldprefixchars = str(conf.supybot.reply.whenAddressedBy.chars)
        nick = 'supybot'
        conf.supybot.reply.whenAddressedBy.chars.set('~!@')
        inChannel = ['~foo', '@foo', '!foo',
                     '%s: foo' % nick, '%s foo' % nick,
                     '%s: foo' % nick.capitalize(), '%s: foo' % nick.upper()]
        inChannel = [ircmsgs.privmsg('#foo', s) for s in inChannel]
        badmsg = ircmsgs.privmsg('#foo', '%s:foo' % nick)
        self.failIf(callbacks.addressed(nick, badmsg))
        badmsg = ircmsgs.privmsg('#foo', '%s^: foo' % nick)
        self.failIf(callbacks.addressed(nick, badmsg))
        for msg in inChannel:
            self.assertEqual('foo', callbacks.addressed(nick, msg), msg)
        msg = ircmsgs.privmsg(nick, 'foo')
        self.assertEqual('foo', callbacks.addressed(nick, msg))
        conf.supybot.reply.whenAddressedBy.chars.set(oldprefixchars)
        msg = ircmsgs.privmsg('#foo', '%s::::: bar' % nick)
        self.assertEqual('bar', callbacks.addressed(nick, msg))
        msg = ircmsgs.privmsg('#foo', '%s: foo' % nick.upper())
        self.assertEqual('foo', callbacks.addressed(nick, msg))
        badmsg = ircmsgs.privmsg('#foo', '%s`: foo' % nick)
        self.failIf(callbacks.addressed(nick, badmsg))

    def testAddressedReplyWhenNotAddressed(self):
        msg1 = ircmsgs.privmsg('#foo', '@bar')
        msg2 = ircmsgs.privmsg('#foo', 'bar')
        self.assertEqual(callbacks.addressed('blah', msg1), 'bar')
        self.assertEqual(callbacks.addressed('blah', msg2), '')
        try:
            original = conf.supybot.reply.whenNotAddressed()
            conf.supybot.reply.whenNotAddressed.setValue(True)
            # need to recreate the msg objects since the old ones have already
            # been tagged
            msg1 = ircmsgs.privmsg('#foo', '@bar')
            msg2 = ircmsgs.privmsg('#foo', 'bar')
            self.assertEqual(callbacks.addressed('blah', msg1), 'bar')
            self.assertEqual(callbacks.addressed('blah', msg2), 'bar')
        finally:
            conf.supybot.reply.whenNotAddressed.setValue(original)

    def testAddressedWithMultipleNicks(self):
        msg = ircmsgs.privmsg('#foo', 'bar: baz')
        self.assertEqual(callbacks.addressed('bar', msg), 'baz')
        # need to recreate the msg objects since the old ones have already
        # been tagged
        msg = ircmsgs.privmsg('#foo', 'bar: baz')
        self.assertEqual(callbacks.addressed('biff', msg, nicks=['bar']),
                         'baz')

    def testAddressedWithNickAtEnd(self):
        msg = ircmsgs.privmsg('#foo', 'baz, bar')
        self.assertEqual(callbacks.addressed('bar', msg,
                                             whenAddressedByNickAtEnd=True),
                         'baz')

    def testAddressedPrefixCharsTakePrecedenceOverNickAtEnd(self):
        msg = ircmsgs.privmsg('#foo', '@echo foo')
        self.assertEqual(callbacks.addressed('foo', msg,
                                             whenAddressedByNickAtEnd=True,
                                             prefixChars='@'),
                         'echo foo')


    def testReply(self):
        prefix = 'foo!bar@baz'
        channelMsg = ircmsgs.privmsg('#foo', 'bar baz', prefix=prefix)
        nonChannelMsg = ircmsgs.privmsg('supybot', 'bar baz', prefix=prefix)
        self.assertEqual(ircmsgs.notice(nonChannelMsg.nick, 'foo'),
                         callbacks.reply(channelMsg, 'foo', private=True))
        self.assertEqual(ircmsgs.notice(nonChannelMsg.nick, 'foo'),
                         callbacks.reply(nonChannelMsg, 'foo'))
        self.assertEqual(ircmsgs.privmsg(channelMsg.args[0],
                                         '%s: foo' % channelMsg.nick),
                         callbacks.reply(channelMsg, 'foo'))
        self.assertEqual(ircmsgs.privmsg(channelMsg.args[0],
                                         'foo'),
                         callbacks.reply(channelMsg, 'foo', prefixNick=False))
        self.assertEqual(ircmsgs.notice(nonChannelMsg.nick, 'foo'),
                         callbacks.reply(channelMsg, 'foo',
                                         notice=True, private=True))

    def testReplyTo(self):
        prefix = 'foo!bar@baz'
        msg = ircmsgs.privmsg('#foo', 'bar baz', prefix=prefix)
        self.assertEqual(callbacks.reply(msg, 'blah', to='blah'),
                         ircmsgs.privmsg('#foo', 'blah: blah'))
        self.assertEqual(callbacks.reply(msg, 'blah', to='blah', private=True),
                         ircmsgs.notice('blah', 'blah'))

    def testTokenize(self):
        self.assertEqual(callbacks.tokenize(''), [])
        self.assertEqual(callbacks.tokenize('foo'), ['foo'])
        self.assertEqual(callbacks.tokenize('foo'), ['foo'])
        self.assertEqual(callbacks.tokenize('bar [baz]'), ['bar', ['baz']])


class AmbiguityTestCase(PluginTestCase):
    plugins = ('Misc',) # Something so it doesn't complain.
    class Foo(callbacks.Plugin):
        def bar(self, irc, msg, args):
            irc.reply('foo.bar')
    class Bar(callbacks.Plugin):
        def bar(self, irc, msg, args):
            irc.reply('bar.bar')

    def testAmbiguityWithCommandSameNameAsPlugin(self):
        self.irc.addCallback(self.Foo(self.irc))
        self.assertResponse('bar', 'foo.bar')
        self.irc.addCallback(self.Bar(self.irc))
        self.assertResponse('bar', 'bar.bar')

class ProperStringificationOfReplyArgs(PluginTestCase):
    plugins = ('Misc',) # Same as above.
    class NonString(callbacks.Plugin):
        def int(self, irc, msg, args):
            irc.reply(1)
    class ExpectsString(callbacks.Plugin):
        def lower(self, irc, msg, args):
            irc.reply(args[0].lower())

    def test(self):
        self.irc.addCallback(self.NonString(self.irc))
        self.irc.addCallback(self.ExpectsString(self.irc))
        self.assertResponse('expectsstring lower [nonstring int]', '1')

## class PrivmsgTestCaseWithKarma(ChannelPluginTestCase):
##     plugins = ('Utilities', 'Misc', 'Web', 'Karma', 'String')
##     conf.allowEval = True
##     timeout = 2
##     def testSecondInvalidCommandRespondsWithThreadedInvalidCommands(self):
##         try:
##             orig = conf.supybot.plugins.Karma.response()
##             conf.supybot.plugins.Karma.response.setValue(True)
##             self.assertNotRegexp('echo [foo++] [foo++]', 'not a valid')
##             _ = self.irc.takeMsg()
##         finally:
##             conf.supybot.plugins.Karma.response.setValue(orig)

    
class PrivmsgTestCase(ChannelPluginTestCase):
    plugins = ('Utilities', 'Misc', 'Web', 'String')
    conf.allowEval = True
    timeout = 2
    def testEmptySquareBrackets(self):
        self.assertError('echo []')

##     def testHelpNoNameError(self):
##         # This will raise a NameError if some dynamic scoping isn't working
##         self.assertNotError('load Http')
##         self.assertHelp('extension')

    def testMaximumNestingDepth(self):
        original = conf.supybot.commands.nested.maximum()
        try:
            conf.supybot.commands.nested.maximum.setValue(3)
            self.assertResponse('echo foo', 'foo')
            self.assertResponse('echo [echo foo]', 'foo')
            self.assertResponse('echo [echo [echo foo]]', 'foo')
            self.assertResponse('echo [echo [echo [echo foo]]]', 'foo')
            self.assertError('echo [echo [echo [echo [echo foo]]]]')
        finally:
            conf.supybot.commands.nested.maximum.setValue(original)

    def testSimpleReply(self):
        self.assertResponse("eval irc.reply('foo')", 'foo')

    def testSimpleReplyAction(self):
        self.assertResponse("eval irc.reply('foo', action=True)",
                            '\x01ACTION foo\x01')

    def testReplyWithNickPrefix(self):
        self.feedMsg('@len foo')
        m = self.irc.takeMsg()
        self.failUnless(m is not None, 'm: %r' % m)
        self.failUnless(m.args[1].startswith(self.nick))
        try:
            original = conf.supybot.reply.withNickPrefix()
            conf.supybot.reply.withNickPrefix.setValue(False)
            self.feedMsg('@len foobar')
            m = self.irc.takeMsg()
            self.failUnless(m is not None)
            self.failIf(m.args[1].startswith(self.nick))
        finally:
            conf.supybot.reply.withNickPrefix.setValue(original)

    def testErrorPrivateKwarg(self):
        try:
            original = conf.supybot.reply.error.inPrivate()
            conf.supybot.reply.error.inPrivate.setValue(False)
            m = self.getMsg("eval irc.error('foo', private=True)")
            self.failUnless(m, 'No message returned.')
            self.failIf(ircutils.isChannel(m.args[0]))
        finally:
            conf.supybot.reply.error.inPrivate.setValue(original)

    def testErrorNoArgumentIsArgumentError(self):
        self.assertHelp('eval irc.error()')

    def testErrorWithNotice(self):
        try:
            original = conf.supybot.reply.error.withNotice()
            conf.supybot.reply.error.withNotice.setValue(True)
            m = self.getMsg("eval irc.error('foo')")
            self.failUnless(m, 'No message returned.')
            self.failUnless(m.command == 'NOTICE')
        finally:
            conf.supybot.reply.error.withNotice.setValue(original)

    def testErrorReplyPrivate(self):
        try:
            original = str(conf.supybot.reply.error.inPrivate)
            conf.supybot.reply.error.inPrivate.set('False')
            # If this doesn't raise an error, we've got a problem, so the next
            # two assertions shouldn't run.  So we first check that what we
            # expect to error actually does so we don't go on a wild goose
            # chase because our command never errored in the first place :)
            s = 're s/foo/bar baz' # will error; should be "re s/foo/bar/ baz"
            self.assertError(s)
            m = self.getMsg(s)
            self.failUnless(ircutils.isChannel(m.args[0]))
            conf.supybot.reply.error.inPrivate.set('True')
            m = self.getMsg(s)
            self.failIf(ircutils.isChannel(m.args[0]))
        finally:
            conf.supybot.reply.error.inPrivate.set(original)

    # Now for stuff not based on the plugins.
    class First(callbacks.Plugin):
        def firstcmd(self, irc, msg, args):
            """First"""
            irc.reply('foo')

    class Second(callbacks.Plugin):
        def secondcmd(self, irc, msg, args):
            """Second"""
            irc.reply('bar')

    class FirstRepeat(callbacks.Plugin):
        def firstcmd(self, irc, msg, args):
            """FirstRepeat"""
            irc.reply('baz')

    class Third(callbacks.Plugin):
        def third(self, irc, msg, args):
            """Third"""
            irc.reply(' '.join(args))

    def tearDown(self):
        if hasattr(self.First, 'first'):
            del self.First.first
        if hasattr(self.Second, 'second'):
            del self.Second.second
        if hasattr(self.FirstRepeat, 'firstrepeat'):
            del self.FirstRepeat.firstrepeat
        ChannelPluginTestCase.tearDown(self)

    def testDispatching(self):
        self.irc.addCallback(self.First(self.irc))
        self.irc.addCallback(self.Second(self.irc))
        self.assertResponse('firstcmd', 'foo')
        self.assertResponse('secondcmd', 'bar')
        self.assertResponse('first firstcmd', 'foo')
        self.assertResponse('second secondcmd', 'bar')
        self.assertRegexp('first first firstcmd',
                'there is no command named "first" in it')

    def testAmbiguousError(self):
        self.irc.addCallback(self.First(self.irc))
        self.assertNotError('firstcmd')
        self.irc.addCallback(self.FirstRepeat(self.irc))
        self.assertError('firstcmd')
        self.assertError('firstcmd [firstcmd]')
        self.assertNotRegexp('firstcmd', '(foo.*baz|baz.*foo)')
        self.assertResponse('first firstcmd', 'foo')
        self.assertResponse('firstrepeat firstcmd', 'baz')

    def testAmbiguousHelpError(self):
        self.irc.addCallback(self.First(self.irc))
        self.irc.addCallback(self.FirstRepeat(self.irc))
        self.assertError('help first')

    def testHelpDispatching(self):
        self.irc.addCallback(self.First(self.irc))
        self.assertHelp('help firstcmd')
        self.assertHelp('help first firstcmd')
        self.irc.addCallback(self.FirstRepeat(self.irc))
        self.assertError('help firstcmd')
        self.assertRegexp('help first firstcmd', 'First', 0) # no re.I flag.
        self.assertRegexp('help firstrepeat firstcmd', 'FirstRepeat', 0)

    class TwoRepliesFirstAction(callbacks.Plugin):
        def testactionreply(self, irc, msg, args):
            irc.reply('foo', action=True)
            irc.reply('bar') # We're going to check that this isn't an action.

    def testNotActionSecondReply(self):
        self.irc.addCallback(self.TwoRepliesFirstAction(self.irc))
        self.assertAction('testactionreply', 'foo')
        m = self.getMsg(' ')
        self.failIf(m.args[1].startswith('\x01ACTION'))

    def testEmptyNest(self):
        try:
            conf.supybot.reply.whenNotCommand.set('True')
            self.assertError('echo []')
            conf.supybot.reply.whenNotCommand.set('False')
            self.assertResponse('echo []', '[]')
        finally:
            conf.supybot.reply.whenNotCommand.set('False')

    def testDispatcherHelp(self):
        self.assertNotRegexp('help first', r'\(dispatcher')
        self.assertNotRegexp('help first', r'%s')

    def testDefaultCommand(self):
        self.irc.addCallback(self.First(self.irc))
        self.irc.addCallback(self.Third(self.irc))
        self.assertError('first blah')
        self.assertResponse('third foo bar baz', 'foo bar baz')

    def testSyntaxErrorNotEscaping(self):
        self.assertError('load [foo')
        self.assertError('load foo]')

    def testNoEscapingAttributeErrorFromTokenizeWithFirstElementList(self):
        self.assertError('[plugin list] list')

    class InvalidCommand(callbacks.Plugin):
        def invalidCommand(self, irc, msg, tokens):
            irc.reply('foo')

    def testInvalidCommandOneReplyOnly(self):
        try:
            original = str(conf.supybot.reply.whenNotCommand)
            conf.supybot.reply.whenNotCommand.set('True')
            self.assertRegexp('asdfjkl', 'not a valid command')
            self.irc.addCallback(self.InvalidCommand(self.irc))
            self.assertResponse('asdfjkl', 'foo')
            self.assertNoResponse(' ', 2)
        finally:
            conf.supybot.reply.whenNotCommand.set(original)

    class BadInvalidCommand(callbacks.Plugin):
        def invalidCommand(self, irc, msg, tokens):
            s = 'This shouldn\'t keep Misc.invalidCommand from being called'
            raise Exception(s)

    def testBadInvalidCommandDoesNotKillAll(self):
        try:
            original = str(conf.supybot.reply.whenNotCommand)
            conf.supybot.reply.whenNotCommand.set('True')
            self.irc.addCallback(self.BadInvalidCommand(self.irc))
            self.assertRegexp('asdfjkl', 'not a valid command',
                    expectException=True)
        finally:
            conf.supybot.reply.whenNotCommand.set(original)


class PluginRegexpTestCase(PluginTestCase):
    plugins = ()
    class PCAR(callbacks.PluginRegexp):
        def test(self, irc, msg, args):
            "<foo>"
            raise callbacks.ArgumentError
    def testNoEscapingArgumentError(self):
        self.irc.addCallback(self.PCAR(self.irc))
        self.assertResponse('test', 'test <foo>')

class RichReplyMethodsTestCase(PluginTestCase):
    plugins = ('Config',)
    class NoCapability(callbacks.Plugin):
        def error(self, irc, msg, args):
            irc.errorNoCapability('admin')
    def testErrorNoCapability(self):
        self.irc.addCallback(self.NoCapability(self.irc))
        self.assertRegexp('error', 'You don\'t have the admin capability')
        self.assertNotError('config capabilities.private admin')
        self.assertRegexp('error', 'Error: You\'re missing some capability')
        self.assertNotError('config capabilities.private ""')


class SourceNestedPluginTestCase(PluginTestCase):
    plugins = ('Utilities',)
    class E(callbacks.Plugin):
        def f(self, irc, msg, args):
            """takes no arguments

            F
            """
            irc.reply('f')

        def empty(self, irc, msg, args):
            pass

        class g(callbacks.Commands):
            def h(self, irc, msg, args):
                """takes no arguments

                H
                """
                irc.reply('h')

            class i(callbacks.Commands):
                def j(self, irc, msg, args):
                    """takes no arguments

                    J
                    """
                    irc.reply('j')

        class same(callbacks.Commands):
            def same(self, irc, msg, args):
                """takes no arguments

                same
                """
                irc.reply('same')

    def test(self):
        cb = self.E(self.irc)
        self.irc.addCallback(cb)
        self.assertEqual(cb.getCommand(['f']), ['f'])
        self.assertEqual(cb.getCommand(['same']), ['same'])
        self.assertEqual(cb.getCommand(['e', 'f']), ['e', 'f'])
        self.assertEqual(cb.getCommand(['e', 'g', 'h']), ['e', 'g', 'h'])
        self.assertEqual(cb.getCommand(['e', 'g', 'i', 'j']),
                                       ['e', 'g', 'i', 'j'])
        self.assertResponse('e f', 'f')
        self.assertResponse('e same', 'same')
        self.assertResponse('e g h', 'h')
        self.assertResponse('e g i j', 'j')
        self.assertHelp('help f')
        self.assertHelp('help empty')
        self.assertHelp('help same')
        self.assertHelp('help e g h')
        self.assertHelp('help e g i j')
        self.assertRegexp('list e', 'f, g h, g i j, and same')

    def testCommandSameNameAsNestedPlugin(self):
        cb = self.E(self.irc)
        self.irc.addCallback(cb)
        self.assertResponse('e f', 'f') # Just to make sure it was loaded.
        self.assertEqual(cb.getCommand(['e', 'same']), ['e', 'same'])
        self.assertResponse('e same', 'same')


class WithPrivateNoticeTestCase(ChannelPluginTestCase):
    plugins = ('Utilities',)
    class WithPrivateNotice(callbacks.Plugin):
        def normal(self, irc, msg, args):
            irc.reply('should be with private notice')
        def explicit(self, irc, msg, args):
            irc.reply('should not be with private notice',
                      private=False, notice=False)
        def implicit(self, irc, msg, args):
            irc.reply('should be with notice due to private',
                      private=True)
    def test(self):
        self.irc.addCallback(self.WithPrivateNotice(self.irc))
        # Check normal behavior.
        m = self.assertNotError('normal')
        self.failIf(m.command == 'NOTICE')
        self.failUnless(ircutils.isChannel(m.args[0]))
        m = self.assertNotError('explicit')
        self.failIf(m.command == 'NOTICE')
        self.failUnless(ircutils.isChannel(m.args[0]))
        # Check abnormal behavior.
        originalInPrivate = conf.supybot.reply.inPrivate()
        originalWithNotice = conf.supybot.reply.withNotice()
        try:
            conf.supybot.reply.inPrivate.setValue(True)
            conf.supybot.reply.withNotice.setValue(True)
            m = self.assertNotError('normal')
            self.failUnless(m.command == 'NOTICE')
            self.failIf(ircutils.isChannel(m.args[0]))
            m = self.assertNotError('explicit')
            self.failIf(m.command == 'NOTICE')
            self.failUnless(ircutils.isChannel(m.args[0]))
        finally:
            conf.supybot.reply.inPrivate.setValue(originalInPrivate)
            conf.supybot.reply.withNotice.setValue(originalWithNotice)
        orig = conf.supybot.reply.withNoticeWhenPrivate()
        try:
            conf.supybot.reply.withNoticeWhenPrivate.setValue(True)
            m = self.assertNotError('implicit')
            self.failUnless(m.command == 'NOTICE')
            self.failIf(ircutils.isChannel(m.args[0]))
            m = self.assertNotError('normal')
            self.failIf(m.command == 'NOTICE')
            self.failUnless(ircutils.isChannel(m.args[0]))
        finally:
            conf.supybot.reply.withNoticeWhenPrivate.setValue(orig)

    def testWithNoticeWhenPrivateNotChannel(self):
        original = conf.supybot.reply.withNoticeWhenPrivate()
        try:
            conf.supybot.reply.withNoticeWhenPrivate.setValue(True)
            m = self.assertNotError("eval irc.reply('y',to='x',private=True)")
            self.failUnless(m.command == 'NOTICE')
            m = self.getMsg(' ')
            m = self.assertNotError("eval irc.reply('y',to='#x',private=True)")
            self.failIf(m.command == 'NOTICE')
        finally:
            conf.supybot.reply.withNoticeWhenPrivate.setValue(original)

class ProxyTestCase(SupyTestCase):
    def testHashing(self):
        msg = ircmsgs.ping('0')
        irc = irclib.Irc('test')
        proxy = callbacks.SimpleProxy(irc, msg)
        # First one way...
        self.failIf(proxy != irc)
        self.failUnless(proxy == irc)
        self.assertEqual(hash(proxy), hash(irc))
        # Then the other!
        self.failIf(irc != proxy)
        self.failUnless(irc == proxy)
        self.assertEqual(hash(irc), hash(proxy))

        # And now dictionaries...
        d = {}
        d[irc] = 'foo'
        self.failUnless(len(d) == 1)
        self.failUnless(d[irc] == 'foo')
        self.failUnless(d[proxy] == 'foo')
        d[proxy] = 'bar'
        self.failUnless(len(d) == 1)
        self.failUnless(d[irc] == 'bar')
        self.failUnless(d[proxy] == 'bar')
        d[irc] = 'foo'
        self.failUnless(len(d) == 1)
        self.failUnless(d[irc] == 'foo')
        self.failUnless(d[proxy] == 'foo')




# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2014, Artur Krysiak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import sys
import unittest
from supybot import questions
from supybot.test import SupyTestCase

if sys.version_info >= (2, 7, 0):
    skipif = unittest.skipIf
else:
    skipif = lambda x, y: lambda z:None

try:
    from unittest import mock  # Python 3.3+
except ImportError:
    try:
        import mock  # Everything else, an external 'mock' library
    except ImportError:
        mock = None

# so complicated construction because I want to
# gain the string 'y' instead of the character 'y'
# the reason of usage this construction is to prove
# that comparing strings by 'is' is wrong
# better solution is usage of '==' operator ;)
_yes_answer = ''.join(['', 'y'])

@skipif(mock is None, 'python-mock is not installed.')
class TestYn(SupyTestCase):
    def test_default_yes_selected(self):
        questions.expect = mock.Mock(return_value=_yes_answer)

        answer = questions.yn('up', default='y')

        self.assertTrue(answer)

    def test_default_no_selected(self):
        questions.expect = mock.Mock(return_value='n')

        answer = questions.yn('up', default='n')

        self.assertFalse(answer)

    def test_yes_selected_without_defaults(self):
        questions.expect = mock.Mock(return_value=_yes_answer)

        answer = questions.yn('up')

        self.assertTrue(answer)

    def test_no_selected_without_defaults(self):
        questions.expect = mock.Mock(return_value='n')

        answer = questions.yn('up')

        self.assertFalse(answer)

    def test_no_selected_with_default_yes(self):
        questions.expect = mock.Mock(return_value='n')

        answer = questions.yn('up', default='y')

        self.assertFalse(answer)

    def test_yes_selected_with_default_yes(self):
        questions.expect = mock.Mock(return_value=_yes_answer)

        answer = questions.yn('up', default='y')

        self.assertTrue(answer)

    def test_yes_selected_with_default_no(self):
        questions.expect = mock.Mock(return_value=_yes_answer)

        answer = questions.yn('up', default='n')

        self.assertTrue(answer)

    def test_no_selected_with_default_no(self):
        questions.expect = mock.Mock(return_value='n')

        answer = questions.yn('up', default='n')

        self.assertFalse(answer)

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2002-2005, Jeremiah Fincher
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import re
import sys
import fnmatch

import supybot.conf as conf
import supybot.gpg as gpg
import supybot.utils as utils
import supybot.ircdb as ircdb
from supybot.commands import *
import supybot.ircutils as ircutils
import supybot.callbacks as callbacks
from supybot.i18n import PluginInternationalization, internationalizeDocstring
_ = PluginInternationalization('User')

class User(callbacks.Plugin):
    """Provides commands for dealing with users, such as registration and
    authentication to the bot. This is a core Supybot plugin that should
    not be removed!"""
    def _checkNotChannel(self, irc, msg, password=' '):
        if password and irc.isChannel(msg.args[0]):
            raise callbacks.Error(conf.supybot.replies.requiresPrivacy())

    @internationalizeDocstring
    def list(self, irc, msg, args, optlist, glob):
        """[--capability=<capability>] [<glob>]

        Returns the valid registered usernames matching <glob>.  If <glob> is
        not given, returns all registered usernames.
        """
        predicates = []
        for (option, arg) in optlist:
            if option == 'capability':
                if arg in conf.supybot.capabilities.private():
                    try:
                        u = ircdb.users.getUser(msg.prefix)
                        if not u._checkCapability('admin'):
                            raise KeyError
                    except KeyError:
                        # Note that it may be raised by checkCapability too.
                        irc.error(_('This is a private capability. Only admins '
                            'can see who has it.'), Raise=True)
                def p(u, cap=arg):
                    try:
                        return u._checkCapability(cap)
                    except KeyError:
                        return False
                predicates.append(p)
        if glob:
            r = re.compile(fnmatch.translate(glob), re.I)
            def p(u):
                return r.match(u.name) is not None
            predicates.append(p)
        users = []
        for u in ircdb.users.values():
            for predicate in predicates:
                if not predicate(u):
                    break
            else:
                users.append(u.name)
        if users:
            utils.sortBy(str.lower, users)
            private = self.registryValue("listInPrivate", msg.args[0])
            irc.reply(format('%L', users), private=private)
        else:
            if predicates:
                irc.reply(_('There are no matching registered users.'))
            else:
                irc.reply(_('There are no registered users.'))
    list = wrap(list, [getopts({'capability':'capability'}),
                       additional('glob')])

    @internationalizeDocstring
    def register(self, irc, msg, args, name, password):
        """<name> <password>

        Registers <name> with the given password <password> and the current
        hostmask of the person registering.  You shouldn't register twice; if
        you're not recognized as a user but you've already registered, use the
        hostmask add command to add another hostmask to your already-registered
        user, or use the identify command to identify just for a session.
        This command (and all other commands that include a password) must be
        sent to the bot privately, not in a channel.
        """
        addHostmask = True
        try:
            ircdb.users.getUserId(name)
            irc.error(_('That name is already assigned to someone.'),
                      Raise=True)
        except KeyError:
            pass
        if ircutils.isUserHostmask(name):
            irc.errorInvalid(_('username'), name,
                             _('Hostmasks are not valid usernames.'),
                             Raise=True)
        try:
            u = ircdb.users.getUser(msg.prefix)
            if u._checkCapability('owner'):
                addHostmask = False
            else:
                irc.error(_('Your hostmask is already registered to %s') % 
                          u.name)
                return
        except KeyError:
            pass
        user = ircdb.users.newUser()
        user.name = name
        user.setPassword(password)
        if addHostmask:
            user.addHostmask(msg.prefix)
        ircdb.users.setUser(user)
        irc.replySuccess()
    register = wrap(register, ['private', 'something', 'something'])

    @internationalizeDocstring
    def unregister(self, irc, msg, args, user, password):
        """<name> [<password>]

        Unregisters <name> from the user database.  If the user giving this
        command is an owner user, the password is not necessary.
        """
        try:
            caller = ircdb.users.getUser(msg.prefix)
            isOwner = caller._checkCapability('owner')
        except KeyError:
            caller = None
            isOwner = False
        if not conf.supybot.databases.users.allowUnregistration():
            if not caller or not isOwner:
                self.log.warning('%s tried to unregister user %s.',
                                 msg.prefix, user.name)
                irc.error(_('This command has been disabled.  You\'ll have to '
                          'ask the owner of this bot to unregister your '
                          'user.'), Raise=True)
        if isOwner or user.checkPassword(password):
            ircdb.users.delUser(user.id)
            irc.replySuccess()
        else:
            irc.error(conf.supybot.replies.incorrectAuthentication())
    unregister = wrap(unregister, ['private', 'otherUser',
                                   additional('anything')])

    @internationalizeDocstring
    def changename(self, irc, msg, args, user, newname, password):
        """<name> <new name> [<password>]

        Changes your current user database name to the new name given.
        <password> is only necessary if the user isn't recognized by hostmask.
        This message must be sent to the bot privately (not on a channel) since
        it may contain a password.
        """
        try:
            id = ircdb.users.getUserId(newname)
            irc.error(format(_('%q is already registered.'), newname))
            return
        except KeyError:
            pass
        if user.checkHostmask(msg.prefix) or user.checkPassword(password):
            user.name = newname
            ircdb.users.setUser(user)
            irc.replySuccess()
    changename = wrap(changename, ['private', 'otherUser', 'something',
                                   additional('something', '')])

    class set(callbacks.Commands):
        @internationalizeDocstring
        def password(self, irc, msg, args, user, password, newpassword):
            """[<name>] <old password> <new password>

            Sets the new password for the user specified by <name> to <new
            password>.  Obviously this message must be sent to the bot
            privately (not in a channel). If the requesting user is an owner
            user (and the user whose password is being changed isn't that same
            owner user), then <old password> needn't be correct.
            """
            try:
                u = ircdb.users.getUser(msg.prefix)
            except KeyError:
                u = None
            if user is None:
                if u is None:
                    irc.errorNotRegistered(Raise=True)
                user = u
            if user.checkPassword(password) or \
               (u and u._checkCapability('owner') and not u == user):
                user.setPassword(newpassword)
                ircdb.users.setUser(user)
                irc.replySuccess()
            else:
                irc.error(conf.supybot.replies.incorrectAuthentication())
        password = wrap(password, ['private', optional('otherUser'),
                                   'something', 'something'])

        @internationalizeDocstring
        def secure(self, irc, msg, args, user, password, value):
            """<password> [<True|False>]

            Sets the secure flag on the user of the person sending the message.
            Requires that the person's hostmask be in the list of hostmasks for
            that user in addition to the password being correct.  When the
            secure flag is set, the user *must* identify before they can be
            recognized.  If a specific True/False value is not given, it
            inverts the current value.
            """
            if value is None:
                value = not user.secure
            if user.checkPassword(password) and \
               user.checkHostmask(msg.prefix, useAuth=False):
                user.secure = value
                ircdb.users.setUser(user)
                irc.reply(_('Secure flag set to %s') % value)
            else:
                irc.error(conf.supybot.replies.incorrectAuthentication())
        secure = wrap(secure, ['private', 'user', 'something',
                               additional('boolean')])

    @internationalizeDocstring
    def username(self, irc, msg, args, hostmask):
        """<hostmask|nick>

        Returns the username of the user specified by <hostmask> or <nick> if
        the user is registered.
        """
        if ircutils.isNick(hostmask):
            try:
                hostmask = irc.state.nickToHostmask(hostmask)
            except KeyError:
                irc.error(_('I haven\'t seen %s.') % hostmask, Raise=True)
        try:
            user = ircdb.users.getUser(hostmask)
            irc.reply(user.name)
        except KeyError:
            irc.error(_('I don\'t know who that is.'))
    username = wrap(username, [first('nick', 'hostmask')])

    class hostmask(callbacks.Commands):
        @internationalizeDocstring
        def hostmask(self, irc, msg, args, nick):
            """[<nick>]

            Returns the hostmask of <nick>.  If <nick> isn't given, return the
            hostmask of the person giving the command.
            """
            if not nick:
                nick = msg.nick
            irc.reply(irc.state.nickToHostmask(nick))
        hostmask = wrap(hostmask, [additional('seenNick')])

        @internationalizeDocstring
        def list(self, irc, msg, args, name):
            """[<name>]

            Returns the hostmasks of the user specified by <name>; if <name>
            isn't specified, returns the hostmasks of the user calling the
            command.
            """
            def getHostmasks(user):
                hostmasks = list(map(repr, user.hostmasks))
                if hostmasks:
                    hostmasks.sort()
                    return format('%L', hostmasks)
                else:
                    return format(_('%s has no registered hostmasks.'),
                                  user.name)
            try:
                user = ircdb.users.getUser(msg.prefix)
                if name:
                    if name != user.name and \
                       not ircdb.checkCapability(msg.prefix, 'owner'):
                        irc.error(_('You may only retrieve your own '
                                  'hostmasks.'), Raise=True)
                    else:
                        try:
                            user = ircdb.users.getUser(name)
                            irc.reply(getHostmasks(user))
                        except KeyError:
                            irc.errorNoUser()
                else:
                    irc.reply(getHostmasks(user))
            except KeyError:
                irc.errorNotRegistered()
        list = wrap(list, ['private', additional('something')])

        @internationalizeDocstring
        def add(self, irc, msg, args, user, hostmask, password):
            """[<name>] [<hostmask>] [<password>]

            Adds the hostmask <hostmask> to the user specified by <name>.  The
            <password> may only be required if the user is not recognized by
            hostmask.  <password> is also not required if an owner user is
            giving the command on behalf of some other user.  If <hostmask> is
            not given, it defaults to your current hostmask.  If <name> is not
            given, it defaults to your currently identified name.  This message
            must be sent to the bot privately (not on a channel) since it may
            contain a password.
            """
            if not hostmask:
                hostmask = msg.prefix
            if not ircutils.isUserHostmask(hostmask):
                irc.errorInvalid(_('hostmask'), hostmask,
                                 _('Make sure your hostmask includes a nick, '
                                 'then an exclamation point (!), then a user, '
                                 'then an at symbol (@), then a host.  Feel '
                                 'free to use wildcards (* and ?, which work '
                                 'just like they do on the command line) in '
                                 'any of these parts.'),
                                 Raise=True)
            try:
                otherId = ircdb.users.getUserId(hostmask)
                if otherId != user.id:
                    irc.error(_('That hostmask is already registered.'),
                              Raise=True)
            except KeyError:
                pass
            if not user.checkPassword(password) and \
               not user.checkHostmask(msg.prefix):
                try:
                    u = ircdb.users.getUser(msg.prefix)
                except KeyError:
                    irc.error(conf.supybot.replies.incorrectAuthentication(),
                              Raise=True)
                if not u._checkCapability('owner'):
                    irc.error(conf.supybot.replies.incorrectAuthentication(),
                              Raise=True)
            try:
                user.addHostmask(hostmask)
            except ValueError as e:
                irc.error(str(e), Raise=True)
            try:
                ircdb.users.setUser(user)
            except ircdb.DuplicateHostmask:
                irc.error(_('That hostmask is already registered.'),
                          Raise=True)
            except ValueError as e:
                irc.error(str(e), Raise=True)
            irc.replySuccess()
        add = wrap(add, ['private', first('otherUser', 'user'),
                         optional('something'), additional('something', '')])

        @internationalizeDocstring
        def remove(self, irc, msg, args, user, hostmask, password):
            """[<name>] [<hostmask>] [<password>]

            Removes the hostmask <hostmask> from the record of the user
            specified by <name>.  If the hostmask given is 'all' then all
            hostmasks will be removed.  The <password> may only be required if
            the user is not recognized by their hostmask.  This message must be
            sent to the bot privately (not on a channel) since it may contain a
            password.  If <hostmask> is
            not given, it defaults to your current hostmask.  If <name> is not
            given, it defaults to your currently identified name.
            """
            if not hostmask:
                hostmask = msg.prefix
            if not user.checkPassword(password) and \
               not user.checkHostmask(msg.prefix):
                u = ircdb.users.getUser(msg.prefix)
                if not u._checkCapability('owner'):
                    irc.error(conf.supybot.replies.incorrectAuthentication())
                    return
            try:
                s = ''
                if hostmask == 'all':
                    user.hostmasks.clear()
                    s = _('All hostmasks removed.')
                else:
                    user.removeHostmask(hostmask)
            except KeyError:
                irc.error(_('There was no such hostmask.'))
                return
            ircdb.users.setUser(user)
            irc.replySuccess(s)
        remove = wrap(remove, ['private', first('otherUser', 'user'),
                               optional('something'), additional('something', '')])

    def callCommand(self, command, irc, msg, *args, **kwargs):
        if command[0] != 'gpg' or \
                (gpg.available and self.registryValue('gpg.enable')):
            return super(User, self) \
                    .callCommand(command, irc, msg, *args, **kwargs)
        else:
            irc.error(_('GPG features are not enabled.'))


    @internationalizeDocstring
    def capabilities(self, irc, msg, args, user):
        """[<name>]

        Returns the capabilities of the user specified by <name>; if <name>
        isn't specified, returns the capabilities of the user calling the
        command.
        """
        try:
            u = ircdb.users.getUser(msg.prefix)
        except KeyError:
            irc.errorNotRegistered()
        else:
            if u == user or u._checkCapability('admin'):
                irc.reply('[%s]' % '; '.join(user.capabilities), private=True)
            else:
                irc.error(conf.supybot.replies.incorrectAuthentication(),
                          Raise=True)
    capabilities = wrap(capabilities, [first('otherUser', 'user')])

    @internationalizeDocstring
    def identify(self, irc, msg, args, user, password):
        """<name> <password>

        Identifies the user as <name>. This command (and all other
        commands that include a password) must be sent to the bot privately,
        not in a channel.
        """
        if user.checkPassword(password):
            try:
                user.addAuth(msg.prefix)
                ircdb.users.setUser(user, flush=False)
                irc.replySuccess()
            except ValueError:
                irc.error(_('Your secure flag is true and your hostmask '
                          'doesn\'t match any of your known hostmasks.'))
        else:
            self.log.warning('Failed identification attempt by %s (password '
                             'did not match for %s).', msg.prefix, user.name)
            irc.error(conf.supybot.replies.incorrectAuthentication())
    identify = wrap(identify, ['private', 'otherUser', 'something'])

    @internationalizeDocstring
    def unidentify(self, irc, msg, args, user):
        """takes no arguments

        Un-identifies you.  Note that this may not result in the desired
        effect of causing the bot not to recognize you anymore, since you may
        have added hostmasks to your user that can cause the bot to continue to
        recognize you.
        """
        user.clearAuth()
        ircdb.users.setUser(user)
        irc.replySuccess(_('If you remain recognized after giving this command, '
                         'you\'re being recognized by hostmask, rather than '
                         'by password.  You must remove whatever hostmask is '
                         'causing you to be recognized in order not to be '
                         'recognized.'))
    unidentify = wrap(unidentify, ['user'])

    @internationalizeDocstring
    def whoami(self, irc, msg, args):
        """takes no arguments

        Returns the name of the user calling the command.
        """
        try:
            user = ircdb.users.getUser(msg.prefix)
            irc.reply(user.name)
        except KeyError:
            irc.reply(_('I don\'t recognize you. You can message me either of these two commands: "user identify <username> <password>" to log in or "user register <username> <password>" to register.'))
    whoami = wrap(whoami)

    @internationalizeDocstring
    def stats(self, irc, msg, args):
        """takes no arguments

        Returns some statistics on the user database.
        """
        users = 0
        owners = 0
        admins = 0
        hostmasks = 0
        for user in ircdb.users.values():
            users += 1
            hostmasks += len(user.hostmasks)
            try:
                if user._checkCapability('owner'):
                    owners += 1
                elif user._checkCapability('admin'):
                    admins += 1
            except KeyError:
                pass
        irc.reply(format(_('I have %s registered users '
                         'with %s registered hostmasks; '
                         '%n and %n.'),
                         users, hostmasks,
                         (owners, 'owner'), (admins, 'admin')))
    stats = wrap(stats)


Class = User

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:




###
# Copyright (c) 2002-2005, Jeremiah Fincher
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import os
import sys
import time
import types
import atexit
import logging
import operator
import textwrap
import traceback

from . import ansi, conf, ircutils, registry, utils
from .utils import minisix

deadlyExceptions = [KeyboardInterrupt, SystemExit]

###
# This is for testing, of course.  Mostly it just disables the firewall code
# so exceptions can propagate.
###
testing = False

class Formatter(logging.Formatter):
    _fmtConf = staticmethod(lambda : conf.supybot.log.format())
    def formatTime(self, record, datefmt=None):
        return timestamp(record.created)

    def formatException(self, exc_info):
        (E, e, tb) = exc_info
        for exn in deadlyExceptions:
            if issubclass(e.__class__, exn):
                raise
        return logging.Formatter.formatException(self, (E, e, tb))

    def format(self, record):
        self._fmt = self._fmtConf()
        if hasattr(self, '_style'): # Python 3
            self._style._fmt = self._fmtConf()
        return logging.Formatter.format(self, record)


class PluginFormatter(Formatter):
    _fmtConf = staticmethod(lambda : conf.supybot.log.plugins.format())


class Logger(logging.Logger):
    def exception(self, *args):
        (E, e, tb) = sys.exc_info()
        tbinfo = traceback.extract_tb(tb)
        path = '[%s]' % '|'.join(map(operator.itemgetter(2), tbinfo))
        eStrId = '%s:%s' % (E, path)
        eId = hex(hash(eStrId) & 0xFFFFF)
        logging.Logger.exception(self, *args)
        self.error('Exception id: %s', eId)
        self.debug('%s', utils.python.collect_extra_debug_data())
        # The traceback should be sufficient if we want it.
        # self.error('Exception string: %s', eStrId)

    def _log(self, level, msg, args, exc_info=None, extra=None):
        msg = format(msg, *args)
        logging.Logger._log(self, level, msg, (), exc_info=exc_info, 
                            extra=extra)


class StdoutStreamHandler(logging.StreamHandler):
    def format(self, record):
        s = logging.StreamHandler.format(self, record)
        if record.levelname != 'ERROR' and conf.supybot.log.stdout.wrap():
            # We check for ERROR there because otherwise, tracebacks (which are
            # already wrapped by Python itself) wrap oddly.
            if not isinstance(record.levelname, minisix.string_types):
                print(record)
                print(record.levelname)
                print(utils.stackTrace())
            prefixLen = len(record.levelname) + 1 # ' '
            s = textwrap.fill(s, width=78, subsequent_indent=' '*prefixLen)
            s.rstrip('\r\n')
        return s

    def emit(self, record):
        if conf.supybot.log.stdout() and not conf.daemonized:
            try:
                logging.StreamHandler.emit(self, record)
            except ValueError: # Raised if sys.stdout is closed.
                self.disable()
                error('Error logging to stdout.  Removing stdout handler.')
                exception('Uncaught exception in StdoutStreamHandler:')

    def disable(self):
        self.setLevel(sys.maxsize) # Just in case.
        _logger.removeHandler(self)
        logging._acquireLock()
        try:
            del logging._handlers[self]
        finally:
            logging._releaseLock()


class BetterFileHandler(logging.FileHandler):
    def emit(self, record):
        msg = self.format(record)
        try:
            self.stream.write(msg)
        except (UnicodeError, TypeError):
            try:
                self.stream.write(msg.encode("utf8"))
            except (UnicodeError, TypeError):
                self.stream.write(msg.encode("utf8").decode('ascii', 'replace'))
        self.stream.write(os.linesep)
        try:
            self.flush()
        except OSError as e:
            if e.args[0] == 28:
                print('No space left on device, cannot flush log.')
            else:
                raise


class ColorizedFormatter(Formatter):
    # This was necessary because these variables aren't defined until later.
    # The staticmethod is necessary because they get treated like methods.
    _fmtConf = staticmethod(lambda : conf.supybot.log.stdout.format())
    def formatException(self, exc_info):
        (E, e, tb) = exc_info
        if conf.supybot.log.stdout.colorized():
            return ''.join([ansi.RED,
                            Formatter.formatException(self, (E, e, tb)),
                            ansi.RESET])
        else:
            return Formatter.formatException(self, (E, e, tb))

    def format(self, record, *args, **kwargs):
        if conf.supybot.log.stdout.colorized():
            color = ''
            if record.levelno == logging.CRITICAL:
                color = ansi.WHITE + ansi.BOLD
            elif record.levelno == logging.ERROR:
                color = ansi.RED
            elif record.levelno == logging.WARNING:
                color = ansi.YELLOW
            if color:
                return ''.join([color,
                                Formatter.format(self, record, *args, **kwargs),
                                ansi.RESET])
            else:
                return Formatter.format(self, record, *args, **kwargs)
        else:
            return Formatter.format(self, record, *args, **kwargs)

conf.registerGlobalValue(conf.supybot.directories, 'log',
    conf.Directory('logs', """Determines what directory the bot will store its
    logfiles in."""))

_logDir = conf.supybot.directories.log()
if not os.path.exists(_logDir):
    os.mkdir(_logDir, 0o755)

pluginLogDir = os.path.join(_logDir, 'plugins')

if not os.path.exists(pluginLogDir):
    os.mkdir(pluginLogDir, 0o755)

try:
    messagesLogFilename = os.path.join(_logDir, 'messages.log')
    _handler = BetterFileHandler(messagesLogFilename)
except EnvironmentError as e:
    raise SystemExit('Error opening messages logfile (%s).  ' \
          'Generally, this is because you are running Supybot in a directory ' \
          'you don\'t have permissions to add files in, or you\'re running ' \
          'Supybot as a different user than you normal do.  The original ' \
          'error was: %s' % (messagesLogFilename, utils.gen.exnToString(e)))

# These are public.
formatter = Formatter('NEVER SEEN; IF YOU SEE THIS, FILE A BUG!')
pluginFormatter = PluginFormatter('NEVER SEEN; IF YOU SEE THIS, FILE A BUG!')

# These are not.
logging.setLoggerClass(Logger)
_logger = logging.getLogger('supybot')
_stdoutHandler = StdoutStreamHandler(sys.stdout)

class ValidLogLevel(registry.String):
    """Invalid log level."""
    handler = None
    minimumLevel = -1
    def set(self, s):
        s = s.upper()
        try:
            try:
                level = logging._levelNames[s]
            except AttributeError:
                level = logging._nameToLevel[s]
        except KeyError:
            try:
                level = int(s)
            except ValueError:
                self.error()
        if level < self.minimumLevel:
            self.error()
        if self.handler is not None:
            self.handler.setLevel(level)
        self.setValue(level)

    def __str__(self):
        # The str() is necessary here; apparently getLevelName returns an
        # integer on occasion.  logging--
        level = str(logging.getLevelName(self.value))
        if level.startswith('Level'):
            level = level.split()[-1]
        return level

class LogLevel(ValidLogLevel):
    """Invalid log level.  Value must be either DEBUG, INFO, WARNING,
    ERROR, or CRITICAL."""
    handler = _handler

class StdoutLogLevel(ValidLogLevel):
    """Invalid log level.  Value must be either DEBUG, INFO, WARNING,
    ERROR, or CRITICAL."""
    handler = _stdoutHandler

conf.registerGroup(conf.supybot, 'log')
conf.registerGlobalValue(conf.supybot.log, 'format',
    registry.String('%(levelname)s %(asctime)s %(name)s %(message)s',
    """Determines what the bot's logging format will be.  The relevant
    documentation on the available formattings is Python's documentation on
    its logging module."""))
conf.registerGlobalValue(conf.supybot.log, 'level',
    LogLevel(logging.INFO, """Determines what the minimum priority level logged
    to file will be.  Do note that this value does not affect the level logged
    to stdout; for that, you should set the value of supybot.log.stdout.level.
    Valid values are DEBUG, INFO, WARNING, ERROR, and CRITICAL, in order of
    increasing priority."""))
conf.registerGlobalValue(conf.supybot.log, 'timestampFormat',
    registry.String('%Y-%m-%dT%H:%M:%S', """Determines the format string for
    timestamps in logfiles.  Refer to the Python documentation for the time
    module to see what formats are accepted. If you set this variable to the
    empty string, times will be logged in a simple seconds-since-epoch
    format."""))

class BooleanRequiredFalseOnWindows(registry.Boolean):
    """Value cannot be true on Windows"""
    def set(self, s):
        registry.Boolean.set(self, s)
        if self.value and os.name == 'nt':
            self.error()

conf.registerGlobalValue(conf.supybot.log, 'stdout',
    registry.Boolean(True, """Determines whether the bot will log to
    stdout."""))
conf.registerGlobalValue(conf.supybot.log.stdout, 'colorized',
    BooleanRequiredFalseOnWindows(False, """Determines whether the bot's logs
    to stdout (if enabled) will be colorized with ANSI color."""))
conf.registerGlobalValue(conf.supybot.log.stdout, 'wrap',
    registry.Boolean(False, """Determines whether the bot will wrap its logs
    when they're output to stdout."""))
conf.registerGlobalValue(conf.supybot.log.stdout, 'format',
    registry.String('%(levelname)s %(asctime)s %(message)s',
    """Determines what the bot's logging format will be.  The relevant
    documentation on the available formattings is Python's documentation on
    its logging module."""))
conf.registerGlobalValue(conf.supybot.log.stdout, 'level',
    StdoutLogLevel(logging.INFO, """Determines what the minimum priority level
    logged will be.  Valid values are DEBUG, INFO, WARNING, ERROR, and
    CRITICAL, in order of increasing priority."""))

conf.registerGroup(conf.supybot.log, 'plugins')
conf.registerGlobalValue(conf.supybot.log.plugins, 'individualLogfiles',
    registry.Boolean(False, """Determines whether the bot will separate plugin
    logs into their own individual logfiles."""))
conf.registerGlobalValue(conf.supybot.log.plugins, 'format',
    registry.String('%(levelname)s %(asctime)s %(message)s',
    """Determines what the bot's logging format will be.  The relevant
    documentation on the available formattings is Python's documentation on
    its logging module."""))


# These just make things easier.
debug = _logger.debug
info = _logger.info
warning = _logger.warning
error = _logger.error
critical = _logger.critical
exception = _logger.exception

# These were just begging to be replaced.
registry.error = error
registry.exception = exception

setLevel = _logger.setLevel

atexit.register(logging.shutdown)

# ircutils will work without this, but it's useful.
ircutils.debug = debug

def getPluginLogger(name):
    if not conf.supybot.log.plugins.individualLogfiles():
        return _logger
    log = logging.getLogger('supybot.plugins.%s' % name)
    if not log.handlers:
        filename = os.path.join(pluginLogDir, '%s.log' % name)
        handler = BetterFileHandler(filename)
        handler.setLevel(-1)
        handler.setFormatter(pluginFormatter)
        log.addHandler(handler)
    if name in sys.modules:
        log.info('Starting log for %s.', name)
    return log

def timestamp(when=None):
    if when is None:
        when = time.time()
    format = conf.supybot.log.timestampFormat()
    t = time.localtime(when)
    if format:
        return time.strftime(format, t)
    else:
        return str(int(time.mktime(t)))

def firewall(f, errorHandler=None):
    def logException(self, s=None):
        if s is None:
            s = 'Uncaught exception'
        if hasattr(self, 'log'):
            logging_function = self.log.exception
        else:
            logging_function = exception
        logging_function('%s in %s.%s:', s, self.__class__.__name__, f.__name__)
    def m(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except Exception:
            if testing:
                raise
            logException(self)
            if errorHandler is not None:
                try:
                    return errorHandler(self, *args, **kwargs)
                except Exception:
                    logException(self, 'Uncaught exception in errorHandler')
    m = utils.python.changeFunctionName(m, f.__name__, f.__doc__)
    return m

class MetaFirewall(type):
    def __new__(cls, name, bases, classdict):
        firewalled = {}
        for base in bases:
            if hasattr(base, '__firewalled__'):
                cls.updateFirewalled(firewalled, base.__firewalled__)
        cls.updateFirewalled(firewalled, classdict.get('__firewalled__', []))
        for (attr, errorHandler) in firewalled.items():
            if attr in classdict:
                classdict[attr] = firewall(classdict[attr], errorHandler)
        return super(MetaFirewall, cls).__new__(cls, name, bases, classdict)

    def getErrorHandler(cls, dictOrTuple, name):
        if isinstance(dictOrTuple, dict):
            return dictOrTuple[name]
        else:
            return None
    getErrorHandler = classmethod(getErrorHandler)

    def updateFirewalled(cls, firewalled, __firewalled__):
        for attr in __firewalled__:
            firewalled[attr] = cls.getErrorHandler(__firewalled__, attr)
    updateFirewalled = classmethod(updateFirewalled)
Firewalled = MetaFirewall('Firewalled', (), {})


class PluginLogFilter(logging.Filter):
    def filter(self, record):
        if conf.supybot.log.plugins.individualLogfiles():
            if record.name.startswith('supybot.plugins'):
                return False
        return True

_handler.setFormatter(formatter)
_handler.addFilter(PluginLogFilter())

_handler.setLevel(conf.supybot.log.level())
_logger.addHandler(_handler)
_logger.setLevel(-1)

_stdoutFormatter = ColorizedFormatter('IF YOU SEE THIS, FILE A BUG!')
_stdoutHandler.setFormatter(_stdoutFormatter)
_stdoutHandler.setLevel(conf.supybot.log.stdout.level())
if not conf.daemonized:
    _logger.addHandler(_stdoutHandler)


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:




###
# Copyright (c) 2005, Jeremiah Fincher
# Copyright (c) 2008-2010, James McCoy
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import supybot.conf as conf
import supybot.registry as registry
from supybot.i18n import PluginInternationalization, internationalizeDocstring
_ = PluginInternationalization('Google')

def configure(advanced):
    from supybot.questions import output, yn
    conf.registerPlugin('Google', True)
    output(_("""The Google plugin has the functionality to watch for URLs
              that match a specific pattern. (We call this a snarfer)
              When supybot sees such a URL, it will parse the web page
              for information and reply with the results."""))
    if yn(_('Do you want the Google search snarfer enabled by default?')):
        conf.supybot.plugins.Google.searchSnarfer.setValue(True)

class Language(registry.OnlySomeStrings):
    transLangs = {'Afrikaans': 'af', 'Albanian': 'sq', 'Amharic': 'am',
                  'Arabic': 'ar', 'Armenian': 'hy', 'Azerbaijani': 'az',
                  'Basque': 'eu', 'Belarusian': 'be', 'Bengali': 'bn',
                  'Bulgarian': 'bg', 'Burmese': 'my', 'Catalan': 'ca',
                  'Chinese': 'zh', 'Chinese_simplified': 'zh-CN',
                  'Chinese_traditional': 'zh-TW', 'Croatian': 'hr',
                  'Czech': 'cs', 'Danish': 'da', 'Dhivehi': 'dv',
                  'Dutch': 'nl', 'English': 'en', 'Esperanto': 'eo',
                  'Estonian': 'et', 'Filipino': 'tl', 'Finnish': 'fi',
                  'French': 'fr', 'Galician': 'gl', 'Georgian': 'ka',
                  'German': 'de', 'Greek': 'el', 'Gujarati': 'gu',
                  'Hebrew': 'iw', 'Hindi': 'hi', 'Hungarian': 'hu',
                  'Icelandic': 'is', 'Indonesian': 'id', 'Inuktitut': 'iu',
                  'Italian': 'it', 'Japanese': 'ja', 'Kannada': 'kn',
                  'Kazakh': 'kk', 'Khmer': 'km', 'Korean': 'ko',
                  'Kurdish': 'ku', 'Kyrgyz': 'ky', 'Laothian': 'lo',
                  'Latvian': 'lv', 'Lithuanian': 'lt', 'Macedonian': 'mk',
                  'Malay': 'ms', 'Malayalam': 'ml', 'Maltese': 'mt',
                  'Marathi': 'mr', 'Mongolian': 'mn', 'Nepali': 'ne',
                  'Norwegian': 'no', 'Oriya': 'or', 'Pashto': 'ps',
                  'Persian': 'fa', 'Polish': 'pl', 'Portuguese': 'pt-PT',
                  'Punjabi': 'pa', 'Romanian': 'ro', 'Russian': 'ru',
                  'Sanskrit': 'sa', 'Serbian': 'sr', 'Sindhi': 'sd',
                  'Sinhalese': 'si', 'Slovak': 'sk', 'Slovenian': 'sl',
                  'Spanish': 'es', 'Swedish': 'sv', 'Tajik': 'tg',
                  'Tamil': 'ta', 'Tagalog': 'tl', 'Telugu': 'te',
                  'Thai': 'th', 'Tibetan': 'bo', 'Turkish': 'tr',
                  'Ukranian': 'uk', 'Urdu': 'ur', 'Uzbek': 'uz',
                  'Uighur': 'ug', 'Vietnamese': 'vi',
                  'Detect language': 'auto'}
    validStrings = ['lang_' + s for s in transLangs.values()]
    validStrings.append('')
    def normalize(self, s):
        if s and not s.startswith('lang_'):
            s = 'lang_' + s
        if not s.endswith('CN') or s.endswith('TW') or s.endswith('PT'):
            s = s.lower()
        else:
            s = s.lower()[:-2] + s[-2:]
        return s

class NumSearchResults(registry.PositiveInteger):
    """Value must be 1 <= n <= 8"""
    def setValue(self, v):
        if v > 8:
            self.error()
        super(self.__class__, self).setValue(v)

class SafeSearch(registry.OnlySomeStrings):
    validStrings = ['active', 'moderate', 'off']

Google = conf.registerPlugin('Google')
conf.registerGlobalValue(Google, 'referer',
    registry.String('', _("""Determines the URL that will be sent to Google for
    the Referer field of the search requests.  If this value is empty, a
    Referer will be generated in the following format:
    http://$server/$botName""")))
conf.registerChannelValue(Google, 'baseUrl',
    registry.String('google.com', _("""Determines the base URL used for
    requests.""")))
conf.registerChannelValue(Google, 'searchSnarfer',
    registry.Boolean(False, _("""Determines whether the search snarfer is
    enabled.  If so, messages (even unaddressed ones) beginning with the word
    'google' will result in the first URL Google returns being sent to the
    channel.""")))
conf.registerChannelValue(Google, 'colorfulFilter',
    registry.Boolean(False, _("""Determines whether the word 'google' in the
    bot's output will be made colorful (like Google's logo).""")))
conf.registerChannelValue(Google, 'bold',
    registry.Boolean(True, _("""Determines whether results are bolded.""")))
conf.registerChannelValue(Google, 'oneToOne',
    registry.Boolean(False, _("""Determines whether results are sent in
    different lines or all in the same one.""")))
conf.registerChannelValue(Google, 'maximumResults',
    NumSearchResults(3, _("""Determines the maximum number of results returned
    from the google command.""")))
conf.registerChannelValue(Google, 'defaultLanguage',
    Language('lang_'+ _('en'), _("""Determines what default language is used in
    searches.  If left empty, no specific language will be requested.""")))
conf.registerChannelValue(Google, 'searchFilter',
    SafeSearch('moderate', _("""Determines what level of search filtering to use
    by default.  'active' - most filtering, 'moderate' - default filtering,
    'off' - no filtering""")))

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:
